/*-----------------------------------------------------------------------------

	ST-Sound ( YM files player library )

	Copyright (C) 1995-1999 Arnaud Carre ( http://leonard.oxg.free.fr )

	YM Music Driver
	
-----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------

	This file is part of ST-Sound

	ST-Sound is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	ST-Sound is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with ST-Sound; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

-----------------------------------------------------------------------------*/

#include <string.h>
#include <stdlib.h>
#include "YmMusic.h"

#define	_LINEAR_OVRS				// Activate linear oversampling (best quality) Only used for DigiMix and UniversalTracker YM file type


// ATARI-ST MFP chip predivisor
static	const int	mfpPrediv[8] = {0,4,10,16,50,64,100,200};



void CYmMusicInit(CYmMusic* ym,int _replayRate)
{
	ym->pBigMalloc = NULL;
	ym->pSongName = NULL;
	ym->pSongAuthor = NULL;
	ym->pSongComment = NULL;
	ym->pSongType = NULL;
	ym->pSongPlayer = NULL;

	ym->pBigSampleBuffer = NULL;
	ym->pMixBlock = NULL;
	
	ym->replayRate = _replayRate;
	ym->innerSamplePos = 0;
	ym->nbDrum = 0;
	ym->pDrumTab = NULL;
	setLoopMode(ym,YM_FALSE);
	CYm2149ExInit(&ym->ymChip,ATARI_CLOCK,1,YM_SMPRATE);
}

void	setTimeControl(CYmMusic* ym,YM_BOOL bTime)
{
		if (bTime)
			ym->attrib |= A_TIMECONTROL;
		else
			ym->attrib &= (~A_TIMECONTROL);
}

void CYmMusicDestroy(CYmMusic* ym)
{
		stop(ym);
		unLoad(ym);
}

void	setLoopMode(CYmMusic* ym,YM_BOOL bLoopMode)
{
		ym->bLoop = bLoopMode;
}

void	setPlayerRate(CYmMusic* ym,int rate)
{
		ym->playerRate = rate;
}

long getPos(CYmMusic* ym)
{
	if (!isSeekable(ym)) return 0;
	if ((ym->nbFrame>0) && (ym->playerRate>0))
	{
		return ((long)ym->currentFrame*1000)/(long)ym->playerRate;
	}
	else
		return 0;

}

long	getMusicTime(CYmMusic* ym)
{
		if ((ym->nbFrame>0) && (ym->playerRate>0))
		{
			return ((long)ym->nbFrame*1000)/(long)ym->playerRate;
		}
		else
			return 0;

}

long	setMusicTime(CYmMusic* ym,long time)
{
		if (!isSeekable(ym)) return 0;
		long newTime = 0;

		if ((ym->songType>=YM_V2) && (ym->songType<YM_VMAX))
		{
			newTime = time;
			if (newTime>=getMusicTime(ym)) newTime = 0;
			ym->currentFrame = (newTime*(long)ym->playerRate)/1000;
		}
		else if ((ym->songType>=YM_TRACKER1) && (ym->songType<YM_TRACKERMAX))
		{
			newTime = time;
			if (newTime>=getMusicTime(ym)) newTime = 0;
			ym->currentFrame = (newTime*(long)ym->playerRate)/1000;
		}

		return newTime;
}

void	getMusicInfo(CYmMusic* ym,ymMusicInfo_t *pInfo)
{
		if (pInfo)
		{
			pInfo->pSongName = ym->pSongName;
			pInfo->pSongAuthor = ym->pSongAuthor;
			pInfo->pSongComment = ym->pSongComment;
			pInfo->pSongType = ym->pSongType;
			pInfo->pSongPlayer = ym->pSongPlayer;

			if (ym->playerRate>0)
				pInfo->musicTimeInSec = (long)ym->nbFrame/(long)ym->playerRate;
			else
				pInfo->musicTimeInSec = 0;
		}
}


void	setAttrib(CYmMusic* ym,int _attrib)
{
		ym->attrib = _attrib;
}

int		getAttrib(CYmMusic* ym)
{
		return ym->attrib;
}

YM_BOOL	isSeekable(CYmMusic* ym)
{
		return getAttrib(ym)&A_TIMECONTROL;
}

void	setLastError(CYmMusic* ym,char *pError)
{
		ym->pLastError = pError;
}

char	*getLastError(CYmMusic* ym)
{
		return ym->pLastError;
}

int		readYmRegister(CYmMusic* ym,int reg)
{
		return readRegister(&ym->ymChip,reg);
}

void	bufferClear(signed short *pBuffer,int nbSample)
{
		memset((void*)pBuffer,0,nbSample*sizeof(signed short));
}

YM_BOOL	update(CYmMusic* ym,YM_SAMPLE *sampleBuffer,int nbSample)
{
int sampleToCompute;
int	vblNbSample;


		if ((!ym->bMusicOk) ||
			(ym->bPause) ||
			(ym->bMusicOver))
		{
			bufferClear(sampleBuffer,nbSample);
			if (ym->bMusicOver)
				return YM_FALSE;
			else
				return YM_TRUE;
		}

		if ((ym->songType >= YM_MIX1) && (ym->songType < YM_MIXMAX))
		{
			stDigitMix(ym,sampleBuffer,nbSample);
		}
		else if ((ym->songType >= YM_TRACKER1) && (ym->songType<YM_TRACKERMAX))
		{
			ymTrackerUpdate(ym,sampleBuffer,nbSample);
		}
		else
		{
			signed short *pOut = sampleBuffer;
			int nbs = nbSample;
			vblNbSample = ym->replayRate/ym->playerRate;
			do
			{
				// Nb de sample  calculer avant l'appel de Player
				sampleToCompute = vblNbSample-ym->innerSamplePos;
				// Test si la fin du buffer arrive avant la fin de sampleToCompute
				if (sampleToCompute>nbs) sampleToCompute = nbs;
				ym->innerSamplePos += sampleToCompute;
				if (ym->innerSamplePos>=vblNbSample)
				{
					player(ym);			// Lecture de la partition (playerRate Hz)
					ym->innerSamplePos -= vblNbSample;
				}
				if (sampleToCompute>0)
				{
					CYm2149ExUpdate(&ym->ymChip,pOut,sampleToCompute);	// YM Emulation.
					pOut += (sampleToCompute<<1);
				}
				nbs -= sampleToCompute;
			}
			while (nbs>0);
		}


		return YM_TRUE;
}



void	readYm6Effect(CYmMusic* ym,unsigned char *pReg,int code,int prediv,int count)
{
int voice;
int ndrum;

		code = pReg[code]&0xf0;
		prediv = (pReg[prediv]>>5)&7;
		count = pReg[count];

		if (code&0x30)
		{
			UD tmpFreq;
			// Ici il y a un effet sur la voie:

			voice = ((code&0x30)>>4)-1;
			switch (code&0xc0)
			{
				case 0x00:		// SID
				case 0x80:		// Sinus-SID

					prediv = mfpPrediv[prediv];
					prediv *= count;
					tmpFreq = 0;
					if (prediv)
					{
						tmpFreq = 2457600L / prediv;
						if ((code&0xc0)==0x00)
							sidStart(&ym->ymChip,voice,tmpFreq,pReg[voice+8]&15);
						else
							sidSinStart(&ym->ymChip,voice,tmpFreq,pReg[voice+8]&15);
					}
					break;

				case 0x40:		// DigiDrum
					ndrum = pReg[voice+8]&31;
					if ((ndrum>=0) && (ndrum<ym->nbDrum))
					{
						int sampleFrq;
						prediv = mfpPrediv[prediv];
						prediv *= count;
						if (prediv>0)
						{
							sampleFrq = 2457600 / prediv;
							drumStart(&ym->ymChip,voice,ym->pDrumTab[ndrum].pData,ym->pDrumTab[ndrum].size,sampleFrq);
						}
					}
					break;

				case 0xc0:		// Sync-Buzzer.

					prediv = mfpPrediv[prediv];
					prediv *= count;
					tmpFreq = 0;
					if (prediv)
					{
						tmpFreq = 2457600L / prediv;
						syncBuzzerStart(&ym->ymChip,tmpFreq,pReg[voice+8]&15);
					}
					break;


			}

		}
}

void	setVolume(CYmMusic* ym,int volume)
{
//		ymChip.setGlobalVolume(volume);
}

void	player(CYmMusic* ym)
 {
 UB	*ptr;
 UD prediv;
 int voice;
 int ndrum;
 int i;


	if (ym->currentFrame<0) ym->currentFrame = 0;

	if (ym->currentFrame>=ym->nbFrame)
	{
		if (ym->bLoop)
		{
			ym->currentFrame = ym->loopFrame;
		}
		else
		{
			ym->bMusicOver = YM_TRUE;
			CYm2149ExReset(&ym->ymChip);
			return;
		}
	}

	ptr = ym->pDataStream+ym->currentFrame*ym->streamInc;

	for (i=0;i<=10;i++)
		writeRegister(&ym->ymChip,i,ptr[i]);

	sidStop(&ym->ymChip,0);
	sidStop(&ym->ymChip,1);
	sidStop(&ym->ymChip,2);
	syncBuzzerStop(&ym->ymChip);

	//---------------------------------------------
	// Check digi-drum
	//---------------------------------------------
	if (ym->songType == YM_V2)		// MADMAX specific !
	{
		if (ptr[13]!=0xff)
		{
			writeRegister(&ym->ymChip,11,ptr[11]);
			writeRegister(&ym->ymChip,12,0);
			writeRegister(&ym->ymChip,13,10);				// MADMAX specific !!
		}
		if (ptr[10]&0x80)					// bit 7 volume canal C pour annoncer une digi-drum madmax.
		{
			int	sampleNum;
			int sampleFrq;
			writeRegister(&ym->ymChip,7,readRegister(&ym->ymChip,7)|0x24)	;	// Coupe TONE + NOISE canal C.
			sampleNum = ptr[10]&0x7f;		// Numero du sample
			sampleFrq = ptr[12];			// Frq sample (valeur envoye au MFP)
			sampleFrq = (2457600>>2) / sampleFrq;	// Frq du digidrum en Hz (le MFP ST tourne  2.4576Mhz)

			drumStart(&ym->ymChip,	2,							// Voice C
								sampleAdress[sampleNum],
								sampleLen[sampleNum],
								sampleFrq);

		}
	}
	else if (ym->songType >= YM_V3)
	{
		writeRegister(&ym->ymChip,11,ptr[11]);
		writeRegister(&ym->ymChip,12,ptr[12]);
		if (ptr[13]!=0xff)
		{
			writeRegister(&ym->ymChip,13,ptr[13]);
		}

		if (ym->songType >= YM_V5)
		{
			int code;

			if (ym->songType == YM_V6)
			{
				readYm6Effect(ym,ptr,1,6,14);
				readYm6Effect(ym,ptr,3,8,15);
			}
			else
			{	// YM5 effect decoding

			//------------------------------------------------------
			// Sid Voice !!
			//------------------------------------------------------
				code = (ptr[1]>>4)&3;
				if (code!=0)
				{
					UD tmpFreq;
					voice = code-1;
					prediv = mfpPrediv[(ptr[6]>>5)&7];
					prediv *= ptr[14];
					tmpFreq = 0;
					if (prediv)
					{
						tmpFreq = 2457600L / prediv;
						sidStart(&ym->ymChip,voice,tmpFreq,ptr[voice+8]&15);
					}
				}

			//------------------------------------------------------
			// YM5 Digi Drum.
			//------------------------------------------------------
				code = (ptr[3]>>4)&3;
				if (code!=0)
				{	// Ici un digidrum demarre sur la voie voice.
					voice = code-1;
					ndrum = ptr[8+voice]&31;
					if ((ndrum>=0) && (ndrum<ym->nbDrum))
					{
						int sampleFrq;
						prediv = mfpPrediv[(ptr[8]>>5)&7];
						prediv *= ptr[15];
						if (prediv)
						{
							sampleFrq = 2457600 / prediv;
							drumStart(&ym->ymChip,voice,ym->pDrumTab[ndrum].pData,ym->pDrumTab[ndrum].size,sampleFrq);
						}
					}
				}
			}
		}
	}
	ym->currentFrame++;
 }

/*

	x x x x x x x x	r0
	0 0 0 0 x x x x	r1		// Special FX 1a
	x x x x x x x x	r2
	0 0 0 0 x x x x r3		// Special FX 2a
	x x x x x x x x r4
	0 0 0 0 x x x x r5
	0 0 0 x x x x x r6		// Special FX 1b
	0 0 x x x x x x r7
	0 0 0 x x x x x r8		// Special FX 2b
	0 0 0 x x x x x r9
	0 0 0 x x x x x r10
	x x x x x x x x r11
	x x x x x x x x r12
	0 0 0 0 x x x x r13
	0 0 0 0 0 0 0 0 r14		// Special FX 1c
	0 0 0 0 0 0 0 0 r15		// Special FX 2c


  Special Fx ?a
	0 0 0 0	: No special FX running
	0 0 0 1 : Sid Voice A
	0 0 1 0 : Sid Voice B
	0 0 1 1 : Sid Voice C
	0 1 0 0 : Extended Fx voice A
	0 1 0 1 : Digidrum voice A
	0 1 1 0 : Digidrum voice B
	0 1 1 1 : Digidrum voice C
	1 0 0 0 : Extended Fx voice B
	1 0 0 1 : Sinus SID voice A
	1 0 1 0 : Sinus SID voice B
	1 0 1 1 : Sinus SID voice C
	1 1 0 0 : Extended Fx voice C
	1 1 0 1 : Sync Buzzer voice A
	1 1 1 0 : Sync Buzzer voice B
	1 1 1 1 : Sync Buzzer voice C



*/

void	readNextBlockInfo(CYmMusic* ym)
{
	ym->nbRepeat--;
	if (ym->nbRepeat<=0)
	{
		ym->mixPos++;
		if (ym->mixPos >= ym->nbMixBlock)
		{
			ym->mixPos = 0;
			if (!ym->bLoop) ym->bMusicOver = TRUE;
		}
		ym->nbRepeat = ym->pMixBlock[ym->mixPos].nbRepeat;
	}
	ym->pCurrentMixSample = ym->pBigSampleBuffer + ym->pMixBlock[ym->mixPos].sampleStart;
	ym->currentSampleLength = (ym->pMixBlock[ym->mixPos].sampleLength)<<12;
	ym->currentPente = (unsigned long)((double)ym->pMixBlock[ym->mixPos].replayFreq*((double)(1<<12)) / (double)YM_SMPRATE);
	ym->currentPos &= ((1<<12)-1);
}

void	stDigitMix(CYmMusic* ym,signed short *pWrite16,int nbs)
{


		if (ym->bMusicOver) return;

		if (ym->mixPos == -1)
		{
			ym->nbRepeat = -1;
			readNextBlockInfo(ym);
		}

		if (nbs) do
		{

			int sa = (int)(signed short)(ym->pCurrentMixSample[ym->currentPos>>12]<<8);
#ifdef _LINEAR_OVRS
			int sb = sa;
			if ((ym->currentPos>>12)<((ym->currentSampleLength>>12)-1))
				sb = (int)(signed short)(ym->pCurrentMixSample[(ym->currentPos>>12)+1]<<8);
			int frac = ym->currentPos&((1<<12)-1);
			sa += (((sb-sa)*frac)>>12);
#endif
			*pWrite16++ = sa;

			ym->currentPos += ym->currentPente;
			if (ym->currentPos>=ym->currentSampleLength)
			{
				readNextBlockInfo(ym);
				if (ym->bMusicOver) return;
			}
		}
		while (--nbs);
}

void	ymTrackerDesInterleave(CYmMusic* ym)
{
unsigned char *a0,*a1,*a2;
unsigned char *pNewBuffer;
int	step;
long int n1,n2;


		if (ym->attrib&A_STREAMINTERLEAVED)
		{
			a0 = ym->pDataStream;
			size_t size = sizeof(ymTrackerLine_t)*ym->nbVoice*ym->nbFrame;
			pNewBuffer = (unsigned char*)malloc(size);
			step = sizeof(ymTrackerLine_t)*ym->nbVoice;
			n1 = step;
			a2 = pNewBuffer;
			do
			{
				n2 = ym->nbFrame;
				a1 = a2;
				do
				{
					*a1 = *a0++;
					a1 += step;
				}
				while (--n2);
				a2++;
			}
			while (--n1);
			memcpy(ym->pDataStream,pNewBuffer,size);
			free(pNewBuffer);
			ym->attrib &= (~A_STREAMINTERLEAVED);
		}
}


void	ymTrackerInit(CYmMusic* ym,int volMaxPercent)
{
int i,s;
int vol;
unsigned long scale;
signed short *pTab;



		for (i=0;i<MAX_VOICE;i++) ym->ymTrackerVoice[i].bRunning = 0;
		ym->ymTrackerNbSampleBefore = 0;

		scale = (256*volMaxPercent) / (ym->nbVoice*100);
		pTab = ym->ymTrackerVolumeTable;

		// Construit la table de volume.
		for (vol=0;vol<64;vol++)
		{
			for (s=-128;s<128;s++)
			{
				*pTab++ = (s*(int)scale*vol)/64;
			}
		}

		// Des-interleave si necessaire.
		ymTrackerDesInterleave(ym);

}


void	ymTrackerPlayer(CYmMusic* ym,ymTrackerVoice_t *pVoice)
{
int i;
ymTrackerLine_t *pLine;


		pLine = (ymTrackerLine_t*)ym->pDataStream;
		pLine += (ym->currentFrame*ym->nbVoice);
	  	for (i=0;i<ym->nbVoice;i++)
		{
			int n;
			pVoice[i].sampleFreq = ((pLine->freq>>8)&255)|((pLine->freq&255)<<8);
			if (pVoice[i].sampleFreq)
			{
				pVoice[i].sampleVolume = pLine->volume&63;
				pVoice[i].bLoop = (pLine->volume&0x40);
				n = pLine->noteOn;
				if (n != 0xff)		// Note ON.
				{
					pVoice[i].bRunning = 1;
					pVoice[i].pSample = ym->pDrumTab[n].pData;
					pVoice[i].sampleSize = ym->pDrumTab[n].size;
					pVoice[i].repLen = ym->pDrumTab[n].repLen;
					pVoice[i].samplePos = 0;
				}
			}
			else
			{
				pVoice[i].bRunning = 0;
			}
			pLine++;
		}

		ym->currentFrame++;
		if (ym->currentFrame >= ym->nbFrame)
		{
			if (!ym->bLoop)
			{
				ym->bMusicOver = YM_TRUE;
			}
			ym->currentFrame = 0;
		}
}


void	ymTrackerVoiceAdd(CYmMusic* ym,ymTrackerVoice_t *pVoice,signed short *pBuffer,int nbs)
{
signed short *pVolumeTab;
unsigned char *pSample;
unsigned long samplePos;
unsigned long sampleEnd;
unsigned long sampleInc;
unsigned long repLen;
double	step;


		if (!(pVoice->bRunning)) return;

		pVolumeTab = &ym->ymTrackerVolumeTable[256*(pVoice->sampleVolume&63)];
		pSample = pVoice->pSample;
		samplePos = pVoice->samplePos;

		step = (double)(pVoice->sampleFreq<<YMTPREC);
		step *= (double)(1<<ym->ymTrackerFreqShift);
		step /= (double)ym->replayRate;
		sampleInc = (unsigned long)step;

		sampleEnd = (pVoice->sampleSize<<YMTPREC);
		repLen = (pVoice->repLen<<YMTPREC);
		if (nbs>0) do
		{
			int va = pVolumeTab[pSample[samplePos>>YMTPREC]];
#ifdef _LINEAR_OVRS
			int vb = va;
			if (samplePos < (sampleEnd-(1<<YMTPREC)))
				vb = pVolumeTab[pSample[(samplePos>>YMTPREC)+1]];
			int frac = samplePos & ((1<<YMTPREC)-1);
			va += (((vb-va)*frac)>>YMTPREC);
#endif
			(*pBuffer++) += va;

			samplePos += sampleInc;
			if (samplePos>=sampleEnd)
			{
				if (pVoice->bLoop)
				{
					samplePos -= repLen;
				}
				else
				{
					pVoice->bRunning = 0;
					return;
				}
			}
		}
		while (--nbs);
		pVoice->samplePos = samplePos;
}

void	ymTrackerUpdate(CYmMusic* ym,signed short *pBuffer,int nbSample)
{
int i;
int _nbs;

		// Clear les buffers.
		memset(pBuffer,0,sizeof(signed short)*nbSample);
		if (ym->bMusicOver) return;

		do
		{
			if (ym->ymTrackerNbSampleBefore == 0)
			{
				// Lit la partition ymTracker
				ymTrackerPlayer(ym,ym->ymTrackerVoice);
				if (ym->bMusicOver) return;
				ym->ymTrackerNbSampleBefore = YMTNBSRATE;
			}
			_nbs = ym->ymTrackerNbSampleBefore;		// nb avant playerUpdate.
			if (_nbs>nbSample) _nbs = nbSample;
			ym->ymTrackerNbSampleBefore -= _nbs;
			if (_nbs>0)
			{
				// Genere les samples.
				for (i=0;i<ym->nbVoice;i++)
				{
					ymTrackerVoiceAdd(ym,&ym->ymTrackerVoice[i],pBuffer,_nbs);
				}
				pBuffer += _nbs;
				nbSample -= _nbs;
			}
		}
		while (nbSample>0);
}
